#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision: 6190 $
$LastChangedDate:: 2022-09-28 11:16:44 +0200#$

This file is part of SyMSpace/ComponentSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import shutil
import stat
import time
import json
import componentspace.osutils.fileutils as fu
from symspace.wrapper.symwrapper import SymProject, SymContainer, SymField, SymDirectory, \
    SymFile, SymMaterial, SymPythonFunction
from componentspace.base_component import Component
from componentspace.utils.image_processing import ImageProcessing
from symspace.osutils.pathutils import Paths
from symspace.osutils.jenkins_utils import isBuildServer
import tempfile
from symspace.wrapper.symwrapper import SymFormula


SymFormula

try:
    # noinspection PyUnresolvedReferences
    from creo.jLink import JLinkInterface
    # noinspection PyUnresolvedReferences
    from magopt.utils import MopFile, ComponentInfo
except Exception as e:
    print(e)


class GeometryComponent(Component):
    """
    Create geometry component definition project
    """

    # ToDo: create a __init__ method which is compatible with the superclass Component
    def __init__(self, geo_file, geo_class=''):
        """
        Init function

        Parameters
        ----------
        geo_file : str
            Geometry file to be converted.

        geo_class : str
            Class of geometry.
        """
        self.geo_file = geo_file
        self.geo_class = geo_class
        self.name = os.path.basename(geo_file).replace('.mop', '_' + geo_class)

        if isBuildServer():
            self.path = os.path.join(tempfile.gettempdir(), self.name)
        else:
            self.path = os.path.join(self.base_directory, self.name)

        self.sym_mdl = None
        self.parCAD = None
        self.lstDrw = {}
        self.lstCADparam = []
        self.lstMATparam = []

    @property
    def base_directory(self):
        return os.path.dirname(self.geo_file)

    @property
    def _geo_class_file(self):
        return os.path.join(self.path, self.name)

    def buildComponent(self, keep_mopc: bool = False):
        print('\n{:_^20}\n'.format(self.geo_class))
        return super().buildComponent(keep_mopc)

    def preBuildTest(self) -> dict:
        """
        Get parameters for test file

        Returns
        -------
        d : dict
        """
        if not os.path.isfile(self.geo_file) and not os.path.isfile(self.geo_file + ".mop"):
            raise FileNotFoundError(self.geo_file + " can not be found.")
        sym_mdl = SymProject.load(self.geo_file)
        if not isinstance(sym_mdl, SymProject):
            raise Exception(self.geo_file + " can not be loaded.")

        webpath = sym_mdl['Info']['WebPath'].getData()
        comp_webpath = '.'.join([webpath, self.geo_class, self.name])

        testparam = sym_mdl['Testparameter']

        sym_mdl.close()

        if testparam is None:
            raise Exception('Container Testparameter does not exist!')

        test_mop_file = os.path.join(self.base_directory, 'tests', 'test_{}'.format(self.name))

        d = {'MopFile': test_mop_file,
             'WebPath': comp_webpath,
             'TestParam': testparam}
        return d

    def createDirectory(self):
        """
        Creates a directory with the name of the geometry and the component class
        Copies directories /rst, /help, and /doc

        """
        # Remove CAD directory
        # print(self.path)
        # print(self.geo_file)
        if os.path.exists(self.path):
            shutil.rmtree(self.path, onerror=fu.on_rm_error)
            # wait until OS is finished
            cnt = 0
            while os.path.exists(self.path):
                time.sleep(1)
                cnt = cnt + 1
                if cnt > 3:
                    raise ValueError('Cannot remove path!')

        # Create CAD directory
        if not os.path.exists(self.path):
            os.mkdir(self.path)

        fname = os.path.join(self.path, self.name + '.mop')
        if os.path.isfile(fname):
            os.remove(fname)
        if os.path.exists(fname + '.Files'):
            shutil.rmtree(fname + '.Files')

        geo_path = os.path.dirname(self.geo_file)
        try:
            fu.copyTree(os.path.join(geo_path, 'rst'), os.path.join(self.path, 'rst'))
        except:
            print('Directory \\rst already exists')

        # Copy src
        try:
            fu.copyTree(os.path.join(geo_path, 'src'), os.path.join(self.path, 'src'))
        except:
            print('Directory \\src already exists')

        try:
            fu.copyTree(os.path.join(geo_path, 'doc'), os.path.join(self.path, 'doc'))
        except:
            print('Directory \\doc already exists')
        # Copy index
        try:
            shutil.copy(os.path.join(geo_path, 'index.rst'), os.path.join(self.path, 'index.rst'))
        except:
            print('Cannot copy index.rst')

        # Copy CADLogo
        src_logo = os.path.join(Paths.getComponentSpaceRoot(), 'Part', 'resources', self.geo_class, 'CADLogo.png')
        dst_logo = os.path.join(self.path, 'rst')

        try:
            if os.path.isfile(os.path.join(dst_logo, 'CADLogo.png')):
                os.remove(os.path.join(dst_logo, 'CADLogo.png'))
            shutil.copy(src_logo, dst_logo)
        except:
            print('Cannot copy logo file')
            print(src_logo)
            print(os.path.join(dst_logo, 'CADLogo.png'))

        # Save source as new file
        prj = SymProject.load(self.geo_file)
        prj.saveAs(self._geo_class_file, True)

    def convertFunction(self):
        """
        Convert container to SymPythonFunction
        """
        self.sym_mdl = SymProject.load(self._geo_class_file)
        self.parCAD = self.sym_mdl.getObject('Project').getParam()[0]

        # Convert Python functions
        fpath = self.path.split('\\')[-1]
        for fnc in self.sym_mdl.getParameters('SymPythonFunction'):
            pyname = os.path.join(fnc.getClassPath(), fnc.getModuleName() + '.py')
            if not os.path.isfile(pyname):
                fnc.setClassPath(fnc.getClassPath().replace('\\' + fpath, ''))
        # self.sym_mdl.regenerate()

        # Add model information
        model = SymField('Model', 'Geometry model')
        self.parCAD.addParam(model, 0)
        model.setData(self.name)
        model.setLock(True)

    def _updateImportAssignmentSlot(self, obj):
        lstImports = obj.getImports()
        for par in lstImports:
            if par.getParamName() == '':
                self.lstCADparam.append(par.getVarName())

        self.lstCADparam = list(dict.fromkeys(self.lstCADparam))

        lstParam = {}
        for par in self.parCAD.getParameters():
            lstParam[par.getName().lower()] = par.getName()

        import_map = {}
        for par in self.lstCADparam:
            if par.lower() in lstParam:
                import_map[par] = lstParam[par.lower()]

        obj.setImportAssignments(import_map)

    def _updateImportAssignment(self):
        lstImports = self.parCAD.getImports()
        for par in lstImports:
            if par.getParamName() == '':
                if par.getVarName().lower().startswith('mat'):
                    self.lstMATparam.append(par.getVarName())
                else:
                    self.lstCADparam.append(par.getVarName())

        self.lstCADparam = list(dict.fromkeys(self.lstCADparam))

        lstParam = {}
        for par in self.parCAD.getParameters():
            lstParam[par.getName().lower()] = par.getName()

        tstParam = {}
        for par in self.sym_mdl['Testparameter'].getParameters('SymParameter'):
            tstParam[par.getName().lower()] = par.getName()

        import_map = {}
        for par in self.lstCADparam:
            if par.lower() in lstParam:
                import_map[par] = lstParam[par.lower()]
            elif par.lower() in tstParam:
                import_map[par] = tstParam[par.lower()]

        for par in self.lstMATparam:
            if 'core' in par.lower():
                if not self.parCAD.getObject('StatorCore') is None:
                    import_map[par] = 'StatorCore.MAT_Core'
                elif not self.parCAD.getObject('RotorCore') is None:
                    import_map[par] = 'RotorCore.MAT_Core'
            elif 'magnet' in par.lower():
                import_map[par] = 'Magnet.MAT_Magnet'
            elif par.lower() in lstParam:
                import_map[par] = lstParam[par.lower()]

        self.parCAD.setImportAssignments(import_map)

    def updateInfo(self):
        """
        Update component info in SymProject
        """
        name = self.sym_mdl.getObject('Name').getData() + '_' + self.geo_class
        self.sym_mdl.getObject('Name').setData(name)
        self.sym_mdl.getObject('Info.DisplayName').setData(name.replace('_', ' ').replace('FreeCAD0p19', 'FreeCAD0.19'))
        self.sym_mdl.getObject('WebPath').setData(self.sym_mdl.getObject('WebPath').getData() + '.' + self.geo_class)

        # Update filter
        # Update provides
        # Removes all filters which do not belong to the current geo_class (e.g. 'FreeCAD' or 'Creo')
        try:
            prov_filter = []
            if self.parCAD.getObject('DRWList') is not None:
                prov_filter.append('FEM_MAG2D')
                for p in self.parCAD.getObject('DRWList').getParam():
                    if 'enable' in p.getName():
                        prov_filter.append(f'FEM_{p.getName().replace("enable", "")}')
            if self.parCAD.getObject('Settings') is not None:
                prov_filter.append('FEM_MAG2D')
                for p in self.parCAD.getObject('Settings').getParam():
                    if 'enable' in p.getName():
                        prov_filter.append(f'FEM_{p.getName().replace("enable", "")}')

            # print(prov_filter)
            for par in self.sym_mdl.getObject('Filter.provides').getParam():
                filter_name = par.getName().replace(f'_{self.geo_class}', '')
                if filter_name not in prov_filter and filter_name.startswith('FEM'):
                    # print(f'------------------------------------------------ Remove {par.getName()}')
                    par.removeParam()
        except Exception as ex:
            print('ERROR setting filter: {}'.format(ex))

        # Update requires
        if self.sym_mdl.getObject('Filter.requires'):
            for rules in self.sym_mdl.getObject('Filter.requires').getParam():
                for flt in rules.getParam():
                    if f'_{self.geo_class}' not in flt.getName():
                        flt.removeParam()

    def createComponent(self, template=None):
        """
        Create Component definition file

        """
        self.createDirectory()
        # self.createBuildScript()
        self.convertFunction()
        self.updateInfo()
        self.sym_mdl.save()

    def generateHelpFiles(self, comp_cont: SymContainer):
        """
        Generate help files

        Help files are copied to the directory above

        Parameters
        ----------
        comp_cont : SymContainer
        """
        super().generateHelpFiles(comp_cont)

        rdir = self.path.split('\\')[-1]
        sdir = os.path.join(self.path, 'rst')
        ddir = sdir.replace('\\' + rdir, '')
        # print(ddir)
        try:
            sfile = os.path.join(sdir, 'Param.txt')
            if os.path.isfile(sfile):
                shutil.copy(sfile, ddir)
        except:
            print('Cannot copy Param.txt')
        try:
            sfile = os.path.join(sdir, 'Paramdetail.txt')
            if os.path.isfile(sfile):
                shutil.copy(sfile, ddir)
        except:
            print('Cannot copy Paramdetail.txt')
        try:
            sfile = os.path.join(sdir, 'ParamIn.txt')
            if os.path.isfile(sfile):
                shutil.copy(sfile, ddir)
        except:
            print('Cannot copy ParamIn.txt')
        try:
            sfile = os.path.join(sdir, 'ParamOut.txt')
            if os.path.isfile(sfile):
                shutil.copy(sfile, ddir)
        except:
            print('Cannot copy ParamOut.txt')

    def writeUnitFile(self):
        """
        Writes Unit file for GeometryPart function
        """
        units = dict()
        file = None
        for inparam in self.parCAD.getImports():
            par = inparam.getHandle()
            if par is not None:
                if hasattr(par, 'getUnit') and par.getUnit() != '':
                    units[par.getName()] = par.getUnit()
                if inparam.getVarName() == 'Part':
                    file = par.getData()
        if file is not None:
            ext = os.path.splitext(file)[1]
            jsonfile = file.replace(ext, '.json')
            with open(jsonfile, 'w') as fp:
                json.dump({'cad': self.geo_class, 'units': units}, fp)


class FreeCADComponent(GeometryComponent):
    """
    Create component file for FreeCAD
    """
    # Todo: move class to a separate file and make sure all references are changed also

    def __init__(self, geo_file, geo_class='FreeCAD', template_file='fnc_geometry_freecad_template.py'):
        super().__init__(geo_file, geo_class)
        self.template_file = template_file

    def createDirectory(self):
        """
        Copy directory /FreeCAD
        """
        super().createDirectory()
        try:
            fu.copyTree(os.path.join(os.path.dirname(self.geo_file), 'FreeCAD'), os.path.join(self.path, 'FreeCAD'))
        except:
            print('Directory \\FreeCAD already exists')

    def convertFunction(self):
        """
        Convert container to SymPythonFunction
        """

        super().convertFunction()

        # Copy _3DModel
        sdir = self.sym_mdl.getObject('FreeCAD.Files.3DModel').getData()
        dir_mdl = os.path.join(self.sym_mdl.getAllFilesDirectory(), '_3DModel')
        fu.copyTree(sdir, dir_mdl)

        # Add 3DModel to component
        model3D = SymDirectory('3DModel')
        self.sym_mdl.getObject('Files').addParam(model3D)
        model3D.setData(dir_mdl)

        freeCADScript = self.sym_mdl.getObject('FreeCAD.FreeCADScript').getData().split('\\')[-1]
        freeCADClass = self.sym_mdl.getObject('FreeCAD.FreeCADClass').getData()

        try:
            freeCADVersion = self.sym_mdl.getObject('FreeCAD.FreeCADVersion').getData()
        except:
            # TODO: Add info message
            freeCADVersion = '0.15'

        # Get Units
        units = ''
        for par in self.parCAD.getParam():
            if isinstance(par, SymField) and par.getUnit() != '':
                units = units + "\n            '{}': '{}',".format(par.getName(), par.getUnit())

        # Copy template script
        dir_script = self.path + '\\FreeCAD'
        if not os.path.exists(dir_script):
            os.mkdir(dir_script)
        python_script = os.path.join(dir_script, 'fnc_freecad_' + freeCADScript)

        fu.copyFileAndReplace(
            os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages', 'lcm_motorbox', 'geometry',
                         self.template_file),
            python_script,
            [('FncGeometryFreecad_', freeCADClass), ('__Script__', freeCADScript),
             ('__Class__', freeCADClass), ('__Version__', freeCADVersion), ('__Units__', units)])

        # Convert to Python function
        self.parCAD.convertToPythonFunction(python_script, freeCADClass)
        self.parCAD = self.sym_mdl.getObject('Project').getParam()[0]
        self.sym_mdl.regenerate()

        # Add Parameters from container FreeCAD.Add to parCAD
        pos = 2
        if self.sym_mdl['FreeCAD.Add'] is not None:
            for par in self.sym_mdl['FreeCAD.Add'].getParam():
                self.parCAD.addParam(par, pos)
                pos += 1
            if self.parCAD['DRWList'] is not None:
                self.parCAD.setImportAssignments({'DRWList': 'DRWList'})

        self.parCAD.setImportAssignments({'Transform': 'Placement'})
        self.parCAD.setExportAssignments({'GeoData': 'GeoData', 'Measurement': 'Measurements'})
        self._updateImportAssignment()

        ''' Slot function '''
        slot_obj = self.parCAD.getObject('Project.Stator.Coil.Slot')
        if slot_obj:
            freeCADClass_slot = freeCADClass + '_slot'
            python_script_slot = os.path.join(dir_script, 'fnc_freecad_slot_' + freeCADScript)

            fu.copyFileAndReplace(
                os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages', 'lcm_motorbox', 'geometry',
                             self.template_file),
                python_script_slot,
                [('FncGeometryFreecad_', freeCADClass_slot), ('__Script__', freeCADScript),
                 ('__Class__', freeCADClass_slot), ('__Version__', freeCADVersion), ('__Units__', units)])
            print(python_script_slot)
            print(freeCADClass_slot)

            slot_obj.convertToPythonFunction(python_script_slot, freeCADClass_slot)
            slot_obj = self.parCAD.getObject('Project.Stator.Coil.Slot')

            if self.parCAD['DRWList'] is not None:
                slot_obj.setImportAssignments({'DRWList': 'DRWList'})

            slot_obj.setImportAssignments({'Transform': 'Placement'})
            slot_obj.setExportAssignments({'GeoData': 'GeoData', 'Measurement': 'Measurements'})

            slot_obj.update()
            self._updateImportAssignmentSlot(slot_obj)

        segmag_obj = self.parCAD.getObject('Project.Rotor.Magnet.Segment')
        if segmag_obj is not None:
            freeCADClass_segmag = freeCADClass + '_segmag'
            python_script_segmag = os.path.join(dir_script, 'fnc_freecad_segmag_' + freeCADScript)

            fu.copyFileAndReplace(
                os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages', 'lcm_motorbox', 'geometry',
                             self.template_file),
                python_script_segmag,
                [('FncGeometryFreecad_', freeCADClass_segmag), ('__Script__', freeCADScript),
                 ('__Class__', freeCADClass_segmag), ('__Version__', freeCADVersion), ('__Units__', units)])

            segmag_obj.convertToPythonFunction(python_script_segmag, freeCADClass_segmag)
            segmag_obj = self.parCAD.getObject('Project.Rotor.Magnet.Segment')

            if self.parCAD['DRWList'] is not None:
                segmag_obj.setImportAssignments({'DRWList': 'DRWList'})

            segmag_obj.setImportAssignments({'Transform': 'Placement'})
            segmag_obj.setExportAssignments({'GeoData': 'GeoData', 'Measurement': 'Measurements'})

            segmag_obj.update()
            self._updateImportAssignmentSlot(segmag_obj)

        self.sym_mdl.regenerate()

    def updateInfo(self):
        super().updateInfo()
        try:
            freeCADVersion = self.sym_mdl.getObject('FreeCAD.FreeCADVersion').getData()
        except:
            freeCADVersion = '0.15'

        self.sym_mdl.getObject('Filter.SimulationTool').setData('FreeCAD_V' + freeCADVersion)


class FreeCAD0p19Component(FreeCADComponent):
    """
    Create component file for FreeCAD 0.19
    """

    # Todo: move class to a separate file and make sure all references are changed also

    def __init__(self, geo_file, geo_class='FreeCAD0p19', template_file='fnc_geometry_freecad_template.py'):
        super().__init__(geo_file, geo_class)
        self.template_file = template_file

    def createDirectory(self):
        """
        Copy directory /FreeCAD
        """
        super().createDirectory()
        try:
            fu.copyTree(os.path.join(os.path.dirname(self.geo_file), 'FreeCAD'), os.path.join(self.path, 'FreeCAD'))
        except:
            print('Directory \\FreeCAD already exists')

    def convertFunction(self):
        """
        Convert container to SymPythonFunction
        """

        super().convertFunction()

        # Copy _3DModel
        sdir = self.sym_mdl.getObject('FreeCAD0p19.Files.3DModel').getData()
        dir_mdl = os.path.join(self.sym_mdl.getAllFilesDirectory(), '_3DModel')
        fu.copyTree(sdir, dir_mdl)

        # Add 3DModel to component
        model3D = SymDirectory('3DModel')
        self.sym_mdl.getObject('Files').addParam(model3D)
        model3D.setData(dir_mdl)

        freeCADScript = self.sym_mdl.getObject('FreeCAD0p19.FreeCADScript').getData().split('\\')[-1]
        freeCADClass = self.sym_mdl.getObject('FreeCAD0p19.FreeCADClass').getData()

        try:
            freeCADVersion = self.sym_mdl.getObject('FreeCAD0p19.FreeCADVersion').getData()
        except:
            # TODO: Add info message
            freeCADVersion = '0.15'

        # Get Units
        units = ''
        for par in self.parCAD.getParam():
            if isinstance(par, SymField) and par.getUnit() != '':
                units = units + "\n            '{}': '{}',".format(par.getName(), par.getUnit())

        # Copy template script
        dir_script = self.path + '\\FreeCAD'
        if not os.path.exists(dir_script):
            os.mkdir(dir_script)
        python_script = os.path.join(dir_script, 'fnc_freecad_' + freeCADScript)

        fu.copyFileAndReplace(
            os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages', 'lcm_motorbox', 'geometry',
                         self.template_file),
            python_script,
            [('FncGeometryFreecad_', freeCADClass), ('__Script__', freeCADScript),
             ('__Class__', freeCADClass), ('__Version__', freeCADVersion), ('__Units__', units)])

        # Convert to Python function
        self.parCAD.convertToPythonFunction(python_script, freeCADClass)
        self.parCAD = self.sym_mdl.getObject('Project').getParam()[0]
        self.sym_mdl.regenerate()

        # Add Parameters from container FreeCAD.Add to parCAD
        pos = 2
        if self.sym_mdl['FreeCAD.Add'] is not None:
            for par in self.sym_mdl['FreeCAD.Add'].getParam():
                self.parCAD.addParam(par, pos)
                pos += 1
            if self.parCAD['DRWList'] is not None:
                self.parCAD.setImportAssignments({'DRWList': 'DRWList'})

        self.parCAD.setImportAssignments({'Transform': 'Placement'})
        self.parCAD.setExportAssignments({'GeoData': 'GeoData', 'Measurement': 'Measurements'})
        self._updateImportAssignment()

        ''' Slot function '''
        slot_obj = self.parCAD.getObject('Project.Stator.Coil.Slot')
        if slot_obj:
            freeCADClass_slot = freeCADClass + '_slot'
            python_script_slot = os.path.join(dir_script, 'fnc_freecad_slot_' + freeCADScript)

            fu.copyFileAndReplace(
                os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages', 'lcm_motorbox', 'geometry',
                             self.template_file),
                python_script_slot,
                [('FncGeometryFreecad_', freeCADClass_slot), ('__Script__', freeCADScript),
                 ('__Class__', freeCADClass_slot), ('__Version__', freeCADVersion), ('__Units__', units)])
            print(python_script_slot)
            print(freeCADClass_slot)
            # raise "asdf"
            slot_obj.convertToPythonFunction(python_script_slot, freeCADClass_slot)
            slot_obj = self.parCAD.getObject('Project.Stator.Coil.Slot')

            if self.parCAD['DRWList'] is not None:
                slot_obj.setImportAssignments({'DRWList': 'DRWList'})

            slot_obj.setImportAssignments({'Transform': 'Placement'})
            slot_obj.setExportAssignments({'GeoData': 'GeoData', 'Measurement': 'Measurements'})

            slot_obj.update()
            self._updateImportAssignmentSlot(slot_obj)

        segmag_obj = self.parCAD.getObject('Project.Rotor.Magnet.Segment')
        if segmag_obj is not None:
            freeCADClass_segmag = freeCADClass + '_segmag'
            python_script_segmag = os.path.join(dir_script, 'fnc_freecad_segmag_' + freeCADScript)

            fu.copyFileAndReplace(
                os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages', 'lcm_motorbox', 'geometry',
                             self.template_file),
                python_script_segmag,
                [('FncGeometryFreecad_', freeCADClass_segmag), ('__Script__', freeCADScript),
                 ('__Class__', freeCADClass_segmag), ('__Version__', freeCADVersion), ('__Units__', units)])

            segmag_obj.convertToPythonFunction(python_script_segmag, freeCADClass_segmag)
            segmag_obj = self.parCAD.getObject('Project.Rotor.Magnet.Segment')

            if self.parCAD['DRWList'] is not None:
                segmag_obj.setImportAssignments({'DRWList': 'DRWList'})

            segmag_obj.setImportAssignments({'Transform': 'Placement'})
            segmag_obj.setExportAssignments({'GeoData': 'GeoData', 'Measurement': 'Measurements'})

            segmag_obj.update()
            self._updateImportAssignmentSlot(segmag_obj)

        self.sym_mdl.regenerate()


class CreoComponent(GeometryComponent):
    """
    Create component file for PTC Creo

    """

    # Todo: move class to a separate file and make sure all references are changed also

    def __init__(self, geo_file, geo_class='Creo'):
        super().__init__(geo_file, geo_class)

    def __del__(self):
        JLinkInterface.exit()

    def createDirectory(self):
        """
        Copy directory /Creo
        """
        super().createDirectory()
        try:
            fu.copyTree(os.path.dirname(self.geo_file) + '\\Creo', self.path + '\\Creo')
        except:
            print('Directory \\Creo already exists')

    def convertFunction(self):
        """
        Convert container to SymPythonFunction
        """

        super().convertFunction()

        # Copy _3DModel
        sdir = self.sym_mdl.getObject('Creo.Files.3DModel').getData()
        dir_mdl = os.path.join(self.sym_mdl.getAllFilesDirectory(), '_3DModel')
        fu.copyTree(sdir, dir_mdl)

        # Add 3DModel to component
        model3D = SymDirectory('3DModel')
        self.sym_mdl.getObject('Files').addParam(model3D)
        model3D.setData(dir_mdl)

        # Create new function from template
        script_name = self.sym_mdl.getObject('Info.Name').getData().lower() + '.py'
        class_name = self.sym_mdl.getObject('Info.Name').getData()

        # Get Units
        units = ''
        for par in self.parCAD.getParam():
            if isinstance(par, SymField) and par.getUnit() != '':
                units = units + "\n            '{}': '{}',".format(par.getName(), par.getUnit())

        # Copy template script
        dir_script = self.path + '\\src'
        if not os.path.exists(dir_script):
            os.mkdir(dir_script)
        python_script = os.path.join(dir_script, 'fnc_creo_' + script_name)

        # Convert parCAD to Python function
        if not isinstance(self.parCAD, SymPythonFunction):
            fu.copyFileAndReplace(
                os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages', 'lcm_motorbox', 'geometry',
                             'fnc_geometry_creo_template.py'),
                python_script,
                [('FncGeometryCreo_', class_name), ('__Units__', units)])

            self.parCAD.convertToPythonFunction(python_script, class_name)
            self.parCAD = self.sym_mdl.getObject('Project').getParam()[0]
        # self.sym_mdl.regenerate()

        # classPath = os.path.join(os.path.join(paths.symspace_home,
        #                                       'Library\\_Geometry\\CreoGeometry\\Functions\\Python\\src\\pygeometry_creo\\fnc_geometry_creo.py'))
        #
        # self.parCAD.convertToPythonFunction(classPath, 'FncGeometryCreo')
        # self.parCAD = self.sym_mdl.getObject('Project').getParam()[0]

        # Add Parameters from container Creo.Add to parCAD
        pos = 2
        for par in self.sym_mdl['Creo.Add'].getParam():
            if isinstance(par, SymFile) and 'Creo' in par.getData():
                # change path to _3DModel
                if not par.getReadOnly():
                    par.setData(os.path.join(dir_mdl, os.path.basename(par.getData())))

            # change path in DRWList
            if isinstance(par, SymMaterial):
                for mdl in par.getParameters('SymFile'):
                    if 'Creo' in mdl.getData():
                        # change path to _3DModel
                        if not mdl.getReadOnly():
                            mdl.setData(os.path.join(dir_mdl, os.path.basename(mdl.getData())))

            self.parCAD.addParam(par, pos)
            pos += 1

        # Replace parameters
        for par in self.sym_mdl['Creo.Replace'].getParam():
            # Find parameter of same type in parCAD
            # ToDo: simplify by using getParamByIdentifier
            lstpar = self.parCAD.getParameters(type(par).__name__)
            try:
                idx = [i.getName() for i in lstpar].index(par.getName())
                container = lstpar[idx].getParent()
                # Find position of parameter
                pos = [i.getName() for i in container.getParam()].index(par.getName())
                lstpar[idx].removeParam()
                container.addParam(par, pos)
            except:
                print(par.getName() + ' not found')

        self.parCAD.setImportAssignments({'DRWList': 'DRWList', 'Transform': 'Placement'})
        self.sym_mdl.regenerate()

        self.parCAD.setImportAssignments({'Assembly': 'Assembly'})
        if self.parCAD.getObject('DRWList') is not None:
            self.lstDrw = self.parCAD.getObject('DRWList').getData()
        else:
            self.lstDrw = dict()
        for key in self.lstDrw.keys():
            self.parCAD.setImportAssignments({key: key})

        # Replace file location for remaining Creo files
        for par in self.sym_mdl.getParameters('SymFile'):
            if os.sep + 'Creo' + os.sep in par.getData() and not par.getReadOnly():
                par.setData(os.path.join(dir_mdl, os.path.basename(par.getData())))

        self.parCAD.setExportAssignments({'GeoData': 'GeoData', 'Measurement': 'Measurements'})
        self.sym_mdl.regenerate()
        self._updateImportAssignment()

        ''' Slot function '''
        slot_obj = self.parCAD.getObject('Project.Stator.Coil.Slot')
        if slot_obj:

            dir_script = self.path + '\\src'
            if not os.path.exists(dir_script):
                os.mkdir(dir_script)
            python_script = os.path.join(dir_script, 'fnc_creo_' + script_name)

            fu.copyFileAndReplace(
                os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages', 'lcm_motorbox', 'geometry',
                             'fnc_geometry_creo_template.py'),
                python_script,
                [('FncGeometryCreo_', class_name), ('__Units__', units)])

            slot_obj.convertToPythonFunction(python_script, class_name)

            slot_obj = self.parCAD.getObject('Project.Stator.Coil.Slot')

            if self.parCAD['DRWList'] is not None:
                slot_obj.setImportAssignments({'DRWList': 'DRWList'})

            slot_obj.setImportAssignments({'Transform': 'Placement'})
            slot_obj.setImportAssignments({'Assembly': 'Assembly'})
            slot_obj.setExportAssignments({'GeoData': 'GeoData', 'Measurement': 'Measurements'})
            self.sym_mdl.regenerate()
            slot_obj.update()
            self._updateImportAssignmentSlot(slot_obj)

        # self._updateImportAssignment()
        self.sym_mdl.regenerate()

    def updateInfo(self):
        super().updateInfo()
        self.sym_mdl.getObject('Filter.SimulationTool').setData('PTC_Creo_V3.0')


class CreoMNFComponent(CreoComponent):
    """
    Create manufacturing component for PTC Creo
    """

    # Todo: move class to a separate file and make sure all references are changed also

    def __init__(self, geo_file, geo_class='Creo_MNF'):
        super().__init__(geo_file, geo_class)

    def createComponent(self, template=None):
        super().createComponent()

        # Rename function
        self.sym_mdl['Info.FunctionName'].setData(self.name)

        # Delete simulation drawings
        if isinstance(self.parCAD['DRWList'], SymContainer):
            for par in self.parCAD['DRWList'].getParam():
                if 'MAG' in par.getName() or \
                        'STRESS' in par.getName() or \
                        'THERM' in par.getName():
                    par.removeParam()
        else:
            delkeys = []
            for key in self.lstDrw.keys():
                if key.startswith('MAG') or key.startswith('STRESS') or key.startswith('THERM'):
                    delkeys.append(key)
            for key in delkeys:
                del self.lstDrw[key]
                if not self.parCAD.getObject(key) is None:
                    self.parCAD.getObject(key).removeParam()
            self.parCAD.getObject('DRWList').setData(self.lstDrw)
            self.sym_mdl.regenerate()

        lstParam = {}
        for par in self.parCAD.getParameters():
            lstParam[par.getName().lower()] = par.getName()

        # Remove parameters from public interface
        pubInterface = {i.lower(): i for i in self.parCAD.listPublicInterfaceName()}
        for par in self.lstCADparam:
            if par.lower() in lstParam:
                self.parCAD.getObject(lstParam[par.lower()]).removeParam()
            # Remove from public interface
            if par.lower() in pubInterface:
                self.parCAD.removePublicInterface(pubInterface[par.lower()])

        # Remove parameters specified in Creo_MNF.delete
        if not self.sym_mdl.getObject('Creo_MNF.Delete') is None:
            for par in self.sym_mdl.getObject('Creo_MNF.Delete').getParameters():
                if not self.parCAD.getObject(par.getName()) is None:
                    self.parCAD.getObject(par.getName()).removeParam()

        # Remove core, coil, magnet and winding
        for par in self.parCAD.getParameters():
            if par.getName() in self.lstDrw:
                continue
            if 'core' in par.getName().lower():
                par.removeParam()
            elif 'coil' in par.getName().lower():
                par.removeParam()
            elif 'magnet' in par.getName().lower():
                par.removeParam()
            elif 'winding' in par.getName().lower():
                par.removeParam()

        # Remove filter provides
        try:
            for par in self.sym_mdl['Filter.provides'].getParam():
                par.removeParam()
        except:
            pass

        try:
            for par in self.sym_mdl['Filter.requires'].getParam():
                par.removeParam()
        except:
            pass

        self.sym_mdl.save()


class CloneFreeCADComponent(FreeCADComponent):
    """
    Clone existing FreeCAD geometry and modify parameter defined in definition project.

    Values of parameters and comments are replaced
    """

    # Todo: move class to a separate file and make sure all references are changed also

    def __init__(self, comp_file, clone_comp):
        super().__init__(clone_comp)
        self.def_mdl = SymProject.load(comp_file)
        self.name = self.def_mdl.getObject('Info.Name').getData()
        if isBuildServer():
            self.path = os.path.join(tempfile.gettempdir(), 'CloneGeometry')
        else:
            self.path = os.path.join(os.path.dirname(comp_file), 'CloneGeometry')
        self.scaleLogo = 0.5

    def createDirectory(self):

        super().createDirectory()

        # Remove rst path in cloned geometry
        rst_path = os.path.join(self.path, 'rst')
        if os.path.exists(rst_path):
            shutil.rmtree(rst_path, onerror=fu.on_rm_error)

        # Copy definition rst to cloned rst
        def_path = self.def_mdl.getDirectory()
        try:
            fu.copyTree(def_path + '\\rst', self.path + '\\rst')
        except:
            print('Directory \\rst already exists')

        # Copy index
        idx_src = os.path.join(def_path, 'index.rst')
        idx_dst = os.path.join(self.path, 'index.rst')
        if os.path.isfile(idx_dst):
            os.chmod(idx_dst, stat.S_IWRITE)
            os.remove(idx_dst)
        try:
            shutil.copy(idx_src, idx_dst)
        except:
            print('Cannot copy index.rst')

    def convertFunction(self):
        """
        Copy param values of definition file to cloned geometry definition
        """
        # ToDo: load model in advance
        sym_mdl = SymProject.load(self._geo_class_file)
        for par in self.def_mdl.getParameters('SymField'):
            mdl_par = sym_mdl.getObject(par.getIdentifier())
            if mdl_par is not None:
                mdl_par.setData(par.getData())
                mdl_par.setComment(par.getComment())

        sym_mdl.save()

        # convertFunction of base class
        super().convertFunction()

        # DocFiles are not generated when Geometry3D.jpg is checked in
        if not os.path.isfile(os.path.join(self.path, 'rst', 'Geometry3D.jpg')):
            self.parCAD.generateDocumentationFiles()
            self.sym_mdl.save()

            img = ImageProcessing(self.parCAD.getDocFile('FrontView')['file'])
            img.strip()

            f_logo = os.path.join(self.path, 'rst', 'logo.png')
            # Add logo if exists
            if os.path.isfile(f_logo):
                logo = ImageProcessing(f_logo)
                logo.strip()
                img.merge(logo)

            img.save(os.path.join(self.path, 'rst', 'Component.png'))

            # Geometry 3D
            img = ImageProcessing(self.parCAD.getDocFile('View3D')['file'])
            img.strip()
            img.save(os.path.join(self.path, 'rst', 'Geometry3D.jpg'))

    def updateInfo(self):
        """
        Update component info in SymProject
        """
        super().updateInfo()
        self.sym_mdl.getObject('Info.Name').setData(self.name)
        self.sym_mdl.getObject('Info.WebPath').setData(self.def_mdl.getObject('WebPath').getData())
