#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 6105 $
$LastChangedDate:: 2022-08-23 09:52:27 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import shutil
import sphinx
import sphinx.cmd.build
import optparse
from bs4 import BeautifulSoup
import symspace.osutils.processutils as pu
from symspace.osutils.pathutils import Paths
from symspace.osutils.pathutils import CMDParser
import componentspace.osutils.fileutils as fu
from componentspace.help.scripts import external_sphinx_process
from componentspace.help.ressources import conf as conf_py


class ComponentHelp:
    # ToDo: make to a subclass of SphinxHelp

    log_dir = os.path.join(Paths.getCSTestReportsDir(), 'component_help')

    def __init__(self, name, component_root="."):
        self.name = name
        self.component_root = os.path.normpath(component_root)

        os.makedirs(self.log_dir, exist_ok=True)

        '''define a sphinx logfile'''
        self.log_file = os.path.join(self.log_dir, 'help_{}.log'.format(name))

    def checkRoot(self):
        """
        check if component root directory contains:
            - index.rst
            - rst folder containing component.png
        """
        if 'index.rst' not in os.listdir(self.component_root):
            print("!!! Error: component root does not contain an index.rst file !!!")
            return 0

        if 'rst' not in os.listdir(self.component_root):
            print("!!! Error: component root does not contain a rst directory !!!")
            return 0

        if 'component.png' not in list(map(str.lower, os.listdir(os.path.join(self.component_root, 'rst')))):
            # e.g. materials don't have component.png!? -> Warning
            print("!!! Warning: no component image (rst/Component.png) found !!!")
            return 2

        return 1

    def copyConfFiles(self):
        """
        copy configuration files from SyMSpace home
        """
        shutil.copyfile(conf_py.__file__, os.path.join(self.component_root, "conf.py"))
        print("... conf.py copied successfully!")

        if not os.path.isdir(os.path.join(self.component_root, "rst", "_static")):
            os.mkdir(os.path.join(self.component_root, "rst", "_static"))
        css_file = os.path.join(Paths.getSyMSpaceRoot(), "Help", "_static", "css", "custom.css")
        shutil.copyfile(css_file, os.path.join(self.component_root, "rst", "_static", "custom.css"))
        print("... custom.css copied successfully!")

    def cleanUpFolders(self):
        """
        clean up unnecessary files,i.e.,
        - only keep .css files from _static folder
        - all javascript and icons are not used for simplified component browser view
        - remove .doctrees directory
        """

        try:
            clean_dir = os.path.join(self.component_root, "help", "_static")
            files = os.listdir(clean_dir)
            for file in files:
                if not file.endswith(".css"):
                    os.remove(os.path.join(clean_dir, file))
            print("... _static folder cleaned up successfully!")
        except Exception as ex:
            print("... _static folder NOT found!!!!!")

        # Remove .Files directories
        try:
            clean_dir = os.path.join(self.component_root, "help")
            files = os.listdir(clean_dir)
            for file in files:
                if file.endswith(".Files") and os.path.exists(os.path.join(clean_dir, file)):
                    shutil.rmtree(os.path.join(clean_dir, file), onerror=fu.on_rm_error)
            print("... .Files folder removed successfully!")
        except Exception as ex:
            print("... .Files NOT found!!!!!")

        try:
            shutil.rmtree(os.path.join(self.component_root, "help", ".doctrees"))
            print("... .doctrees folder cleaned up successfully!")
        except Exception as ex:
            print("... .doctrees NOT found!!!!!")

        files = os.listdir(os.path.join(self.component_root, "help"))
        for file in files:
            if not file == "index.html":
                if os.path.isfile(os.path.join(self.component_root, "help", file)):
                    os.remove(os.path.join(self.component_root, "help", file))
        print("... root directory cleaned up successfully!")

        if os.path.isfile(os.path.join(self.component_root, 'conf.py')):
            os.remove(os.path.join(self.component_root, 'conf.py'))

    # Todo: remove "writeCleanHtml" if it is not used anymore
    def writeCleanHtml(self):
        """
        write new help.html based on index.html but removing all unnecessary parts
        """
        file = open(os.path.join(self.component_root, "help", "index.html"))
        soup = BeautifulSoup(file, "html.parser")

        """
        alles erforderliche auslesen und neue Datei basteln
        """
        math_script = None
        # <head> : only keep MathJax script
        for script in soup.find_all("script", attrs={"type": "text/javascript"}):
            if "mathjax" in str(script):
                math_script = script

        # stylesheets
        style_links = soup.find_all("link", attrs={"rel": "stylesheet"})

        # main body div element
        main_div = soup.find("div", attrs={"class": "body", "role": "main"})

        # write new file
        help_file = open(os.path.join(self.component_root, "help", "help.html"), 'w')

        help_file.write(
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 '
            'Transitional//EN"\n"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">')
        help_file.write('\n<html xmlns="http://www.w3.org/1999/xhtml">')
        help_file.write('\n\t<head>')
        help_file.write('\n\t\t<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />')

        for style in style_links:
            help_file.write('\n\t\t' + str(style))
        if math_script is not None:
            help_file.write('\n\t\t' + str(math_script))
        help_file.write('\n\t</head>')
        help_file.write('\n\t<body>')

        # main_div
        # help_file.write('\n\t\t' + str(main_div).replace('\n', '\n\t\t'))
        help_file.write(r'\n\t\t' + str(main_div).replace('\n', '\n\t\t').replace('\xa0', '&#160;'))

        # '''add footer'''
        # help_file.write('\n\t\t<footer>')
        # help_file.write('\n\t\tVersion X.Y.Z')
        # help_file.write('\n\t\t</footer>')

        help_file.write('\n\t</body>')
        help_file.write('\n</html>')

        file.close()
        help_file.close()
        print("... clean help.html file written successfully!")

        os.remove(file.name)
        print("... original index.html file removed!")

        shutil.move(os.path.join(self.component_root, "help", "help.html"),
                    os.path.join(self.component_root, "help", "index.html"))

    def copyComponentImage(self):
        """
        copy component image component.png used for component browser preview
        """
        files = os.listdir(os.path.join(self.component_root, "rst"))
        component_image_found = False
        for file in files:
            if file.lower() == 'component.png':
                shutil.copyfile(os.path.join(self.component_root, "rst", file),
                                os.path.join(self.component_root, "help", "Component.png"))
                print("... {} successfully copied!".format(file))
                component_image_found = True
        if not component_image_found:
            print("\n!!! WARNING: no component image found !!!\n")

    def build(self):
        """
        Run the sphinx build
        """

        try:
            self.buildInternalProcess()
            if not os.path.isfile(self.log_file):
                raise FileExistsError('Internal Process: No such file {}'.format(self.log_file))
        except AssertionError:
            self.buildExternalProcess()
            if not os.path.isfile(self.log_file):
                raise FileExistsError('External Process: No such file {}'.format(self.log_file))

        self.analyzeLogFile()

    def analyzeLogFile(self):
        with open(self.log_file, 'rb') as fh:
            '''read log file'''
            lines = fh.readlines()

        for line in lines:
            for key in [b'warning', b'error']:
                if key in line.lower():
                    if b'Explicit markup ends without a blank line' in line:
                        print('... suppressed warning: Explicit markup ends without a blank line')
                    elif b'The "csv-table" directive requires content;' in line:
                        print('... suppressed warning: The "csv-table" directive requires content')
                    elif b'is already registered, its visitors will be overridden' in line:
                        print('... suppressed warning: is already registered, its visitors will be overridden')
                    elif b'Unknown interpreted text role "cite:p"' in line:
                        print('... suppressed warning: Unknown interpreted text role "cite:p"')
                    elif b'Unknown directive type "bibliography"' in line:
                        print('... suppressed warning: Unknown directive type "bibliography"')
                    elif b'toctree references unknown document' in line:
                        print('... suppressed warning: toctree references unknown document')
                    elif b'toctree contains reference to nonexisting document' in line:
                        print('... suppressed warning: toctree contains reference to nonexisting document')
                    else:
                        raise Exception('Sphinx build Error:\n {}'.format(line))

    def buildInternalProcess(self):
        """
        Run the sphinx build in an internal process
        """
        source_dir = self.component_root
        output_dir = os.path.join(source_dir, 'help')
        build_options = [source_dir, output_dir, '-b', 'html']
        # build_options = [source_dir, output_dir, '-b', 'html', '-D', 'html_title={}'.format(self.name)]

        '''create logfile'''
        build_options.extend(['-w', self.log_file])

        '''run in nit-picky mode,'''
        build_options.extend(['-n'])

        sphinx.cmd.build.build_main(build_options)

    def buildExternalProcess(self):
        """
        Run the sphinx build in an external process

        Prevents problems due to SyMSpace
        """
        source_dir = self.component_root
        output_dir = os.path.join(source_dir, 'help')

        script_py = external_sphinx_process.__file__
        cmd_args = [source_dir, output_dir, self.log_file]

        try:
            pu.pythonRun(script_py, cmd_args=cmd_args)
        except TimeoutError as ext:
            print('Error in External Sphinx Process: {}'.format(ext))
        except Exception as ex:
            raise Exception('Error in External Sphinx Process: ' + str(ex))

    @classmethod
    def fromCMDParser(cls) -> 'ComponentHelp':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------

        --name : str
        --path : str

        Returns
        -------
        comp_help : ComponentHelp
        """

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("-n", "--name", dest="comp_name",
                          default=0, type="string",
                          help="Component name")
        parser.add_option("-d", "--dir", dest="comp_dir", default=Paths.getComponentSpaceRoot(),
                          type="string", help="Component directory.")
        args = CMDParser.jepSysArgs()
        (options, args) = parser.parse_args(args)
        if len(args) != 0:
            parser.error("incorrect number of arguments")

        comp_name = options.comp_name
        comp_dir = CMDParser.jepArgPath(options.comp_dir)

        comp_help = cls(comp_name, comp_dir)
        return comp_help
